\ rpi-gpio.fth

((
Copyright (c) 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

See:
  http://elinux.org/RPi_Low-level_peripherals

To do
=====

Change history
==============
20140816 SFP001 First cut
))


only forth definitions decimal

: map-io	\ caddr len -- ptr|0
\ *G map the given address range into the Forth address space. The base
\ ** address used by Forth is returned on success. On failure,
\ ** the return value is zero. You must have *\b{root} privileges
\ ** to run *\fo{map-io}, usually by using *\b{sudo} to run the application.
  {: caddr len | fid -- xxx :}
  s" /dev/mem" O_RDWR O_SYNC or open-file
  if  drop 0 exit  endif
  -> fid

  caddr len 4095 + -4096 and		\ min 1 page
  PROT_READ PROT_WRITE or
  MAP_SHARED fid caddr 
  mmap dup MAP_FAILED =
  if  drop 0  endif

  fid close-file drop
;

: unmap-io	\ caddr len --
\ *G Release the mapping set by a previous *\fo{map-io}.
  munmap drop
;
  

4096 constant /mempage	\ -- len
\ *G Memory page size.

$2000:0000 constant _PERIPHS		\ -- addr
\ *G Base address of all BCM2835 on-chip peripherals.

_PERIPHS $0020:0000 + constant _GPIO	\ -- addr
\ *G Base address of on-chip GPIO peripheral.

\ *P GPIO register offsets
\ *[
$000 constant GPFSEL0
$004 constant GPFSEL1
$008 constant GPFSEL2
$00C constant GPFSEL3
$010 constant GPFSEL4
$014 constant GPFSEL5

$01C constant GPSET0
$020 constant GPSET1

$028 constant GPCLR0
$02C constant GPCLR1

$034 constant GPLEV0
$038 constant GPLEV1

$040 constant GPEDS0
$044 constant GPEDS1

$04C constant GPREN0
$050 constant GPREN1

$058 constant GPFEN0
$05C constant GPFEN1

$064 constant GPHEN0
$068 constant GPHEN1

$070 constant GPLEN0
$074 constant GPLEN1

$07C constant GPAREN0
$080 constant GPAREN1

$088 constant GPAFEN0
$08C constant GPAFEN1

$094 constant GPPUD
$098 constant GPPUDCLK0
$09C constant GPPUDCLK1
\ *]

0 value *gpio	\ -- addr
\ *G Returns the mapped address of the GPIO port. 

[undefined] setMask [if]
: setMask	\ value mask addr --
\ *G Clear the *\i{mask} bits at *\i{addr} and set (or) the
\ ** bits defined by *\i{value}.
  tuck @				\ -- value addr mask x
  swap invert and			\ -- value addr x'
  rot or				\ -- addr x''
  swap !
;
[then]

: >mask/off	\ fun# gpio# -- fun#' mask offset
  #10 /mod 4 * -rot 3 *			\ -- offset fun# cnt
					\ 10 GPIOs per reg, 3 bits per GPIO
  tuck lshift  $07 rot lshift  rot	\ -- fun#' mask offset
;
 
: isPinFunction	\ fun# gpio# --
\ *G Set the given GPIO pin (0..53) to the given alternate function.
  >mask/off GPFSEL0 + *gpio + setMask
;

\ *P The alternate function numbers.
\ *[
$04 constant AF0	\ -- x
$05 constant AF1	\ -- x
$06 constant AF2	\ -- x
$07 constant AF3	\ -- x
$03 constant AF4	\ -- x
$02 constant AF5	\ -- x
\ *]

: isInput	\ gpio# --
\ *G Set the given GPIO pin (0..53) to be an input.
  0 swap isPinFunction
;

: isOutput	\ gpio# --
\ *G Set the given GPIO pin (0..53) to be an output.
   1 swap isPinFunction
;
  
: setPin	\ gpio# --
\ *G Set the given GPIO pin number.
  dup $1F and 1 swap lshift		\ -- gpio# mask
  swap 5 rshift 2 lshift  		\ -- mask offset
  GPSET0 + *gpio + !			\ write to set register
;

: clrPin	\ gpio# --
\ *G Set the given GPIO pin number.
  dup $1F and 1 swap lshift		\ -- gpio# mask
  swap 5 rshift 2 lshift  		\ -- mask offset
  GPCLR0 + *gpio + !			\ write to clear register
;

: getPin	\ gpio# -- 0/1
\ *G Read the given GPIO pin number.
  dup $1F and 1 swap lshift		\ -- gpio# mask
  swap 5 rshift 2 lshift  		\ -- mask offset
  GPLEV0 + *gpio + @			\ read level register
  and 0<> 1 and
;

: init-gpio	\ --
\ *G Initialise the GPIO system. After using this word successfully,
\ ** *\fo{*gpio} will return non-zero.
  _GPIO /mempage map-io -> *gpio
;

: term-gpio	\ --
\ *G Shut down the GPIO system.
  _GPIO /mempage unmap-io  0 -> *gpio
;

